#!/usr/bin/env python
# coding=utf-8
"""Utilities for timing, logging, etc."""

import timeit
import re
import sys
import os
import datetime
import argparse


# Timing code, similar to MATLAB's tic/toc
__tictoc_timers = {}


def tic(label='default'):
    """Start a timer."""
    __tictoc_timers[label] = timeit.default_timer()


def toc(label='default', seconds=False):
    """Stop timer and print elapsed time."""
    if label in __tictoc_timers:
        elapsed = timeit.default_timer() - __tictoc_timers[label]
        if seconds:
            print('[{:s}] Elapsed time: {:.3f} s'.format(label, elapsed))
        else:
            print('[{:s}] Elapsed time: {:.2f} ms'.format(label, 1000.0*elapsed))


def ttoc(label='default', seconds=False):
    """Stop timer and return elapsed time."""
    if label in __tictoc_timers:
        elapsed = timeit.default_timer() - __tictoc_timers[label]
        if seconds:
            return elapsed
        else:
            return 1000.0*elapsed


def toc_nsec(label='default', nsec=0.5, seconds=False):
    """Stop timer and print elapsed time (mute output for nsec seconds)."""
    if label in __tictoc_timers:
        elapsed = timeit.default_timer() - __tictoc_timers[label]
        if seconds:
            s = '[{:s}] Elapsed time: {:.3f} s'.format(label, elapsed)
        else:
            s = '[{:s}] Elapsed time: {:.2f} ms'.format(label, 1000.0*elapsed)
        log_nsec(s, nsec, label)


# Log only once every x sec
__log_timers = {}


def log_nsec(string, nsec, label='default'):
    """Display 'string' only once every nsec seconds (floating point number). Use it to avoid spamming your terminal."""
    if label in __log_timers:
        elapsed = timeit.default_timer() - __log_timers[label]
        if elapsed < nsec:
            return
    print(string)
    __log_timers[label] = timeit.default_timer()


# Math
# def rand_mod(m):
#    """Correctly sample a random number modulo m (avoiding modulo bias)"""
#    # python's random lib has random.uniform(a,b), a <= N <= b
#    return random.uniform(0, m-1)
#
# Problem in C/C++:
#   rand() returns a number in [0, RAND_MAX], assume RAND_MAX=10, we want mod 3:
#   rand() = 0, 3, 6, 9;  then mod3 = 0; prob(0) = 4/11
#   rand() = 1, 4, 7, 10; then mod3 = 1; prob(1) = 4/11
#   rand() = 2, 5, 8; then mod3 = 2; prob(2) = 3/11 !!!
#  see also: https://stackoverflow.com/a/10984975/400948


def compare(a, b):
    """Python3 replacement for Python 2.x cmp(), see
    https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons
    """
    return (a > b) - (a < b)


def compare_version_strings(v1, v2):
    """Compares version strings, returns -1/0/+1 if v1 less, equal or greater v2"""
    # Based on https://stackoverflow.com/a/1714190/400948
    def normalize_version_string(v):
        return [int(x) for x in re.sub(r'(\.0+)*$', '', v).split(".")]
    return compare(normalize_version_string(v1), normalize_version_string(v2))


# Make unicode strings, works for Python 2 & 3
try:
    to_unicode = unicode
except NameError:
    to_unicode = str


def slugify(s):
    """Converts a string to a slug (strip special characters,
    replace white space, convert to lowercase...) to be used for file names or
    URLs."""
    import unicodedata
    s = unicodedata.normalize('NFKD', to_unicode(s)).encode('ascii', 'ignore').decode('ascii')
    s = to_unicode(re.sub(r'[^\w\s-]', '', s).strip().lower())
    s = to_unicode(re.sub(r'[-\s]+', '-', s))
    return s


def find_first_index(l, x):
    """Returns the first index of element x within the list l."""
    for idx in range(len(l)):
        if l[idx] == x:
            return idx
    raise ValueError("'{}' is not in list".format(x))


def find_last_index(l, x):
    """Returns the last index of element x within the list l"""
    for idx in reversed(range(len(l))):
        if l[idx] == x:
            return idx
    raise ValueError("'{}' is not in list".format(x))


def argsort(seq, indices_only=False):
    """Returns the sorted indices and the sorted array (seq) if indices_only=False."""
    if indices_only:
        return sorted(range(len(seq)), key=seq.__getitem__)
    else:
        from operator import itemgetter
        return zip(*sorted(enumerate(seq), key=itemgetter(1)))


def is_tool(name):
    """Check whether `name` is on PATH and marked as executable."""
    if sys.version_info >= (3, 3):
        # Taken from https://stackoverflow.com/a/34177358
        from shutil import which
        return which(name) is not None
    else:
        # Search the PATH variable, taken from https://stackoverflow.com/a/5227009
        for path in os.environ['PATH'].split(os.pathsep):
            if os.path.exists(os.path.join(path, name)):
                return True
        return False


def date_str(delimiter=['', '', '-', '', ''], dt=None):
    """Returns a YYYY*MM*DD*hh*mm*ss string using the given delimiters.
    Provide less delimiter to return shorter strings, e.g.
    delimiter=['-'] returns YYYY-MM
    delimiter=['',''] returns YYYYMMDD
    etc.

    Useful to generate timestamped output folder/file names.
    You can pass a custom datetime.datetime object dt. If dt is None,
    datetime.datetime.now() will be taken
    """
    if dt is None:
        now = datetime.datetime.now()
    else:
        now = dt
    res_str = now.strftime('%Y')
    month = now.strftime('%m')
    day = now.strftime('%d')
    hour = now.strftime('%H')
    minute = now.strftime('%M')
    sec = now.strftime('%S')
    num_delim = len(delimiter)
    if num_delim == 0:
        return res_str
    res_str += '{:s}{:s}'.format(delimiter[0], month)
    if num_delim == 1:
        return res_str
    res_str += '{:s}{:s}'.format(delimiter[1], day)
    if num_delim == 2:
        return res_str
    res_str += '{:s}{:s}'.format(delimiter[2], hour)
    if num_delim == 3:
        return res_str
    res_str += '{:s}{:s}'.format(delimiter[3], minute)
    if num_delim == 4:
        return res_str
    res_str += '{:s}{:s}'.format(delimiter[4], sec)
    if num_delim > 5:
        raise RuntimeError('Too many delimiter, currently we only support formating up until seconds')
    return res_str


################################################################################
# Data validation (e.g. argument parsing)

def check_positive_int(value):
    iv = int(value)
    if iv <= 0:
        raise argparse.ArgumentTypeError("%s must be > 0" % value)
    return iv


def check_positive_real(value):
    fv = float(value)
    if fv <= 0:
        raise argparse.ArgumentTypeError("%s must be > 0.0" % value)
    return fv



#!/usr/bin/env python
# coding=utf-8
"""Utilities you'll often need when working with images ;-)"""

import io
import numpy as np
from PIL import Image


def flip_layers(nparray):
    """
    Flip RGB to BGR image data (numpy ndarray).
    Also accepts rgbA/bgrA and single channel images without crashing.
    """
    if len(nparray.shape) == 3:
        if nparray.shape[2] == 4:
            # We got xyzA, make zyxA
            return nparray[..., [2, 1, 0, 3]]
        else:
            return nparray[:, :, ::-1]
    return nparray


def imread(filename, mode='RGB', flip_channels=False):
    """Load an image (using PIL) into a NumPy array.

    Optionally specify PIL's loading 'mode', i.e. 'RGB' for color, 'RGBA' for a
    transparent image and 'L' for grayscale. You can also flip the channels,
    i.e. convert RGB to BGR if you need it."""
    if filename is None:
        return None
    image = np.asarray(Image.open(filename).convert(mode))
    if flip_channels:
        return flip_layers(image)
    else:
        return image


try:
    # Try to load OpenCV (in case you installed it in your workspace)
    import cv2

    def imsave(filename, image, flip_channels=False):
        """Store an image using OpenCV, optionally flipping layers, i.e. BGR -> RGB."""
        if flip_channels:
            cv2.imwrite(filename, flip_layers(image))
        else:
            cv2.imwrite(filename, image)
except:
    # Fall back to Pillow
    def imsave(filename, image, flip_channels=False):
        """Store an image using PIL, optionally flipping layers, i.e. BGR -> RGB."""
        if flip_channels:
            im_np = flip_layers(image)
        else:
            im_np = image
        # TODO On my Ubuntu 16.04 machine, PIL's JPEG preset were crap. Need
        # to check this for other versions.
        Image.fromarray(im_np).save(filename)


def ndarray2memory_file(np_data, format='png'):
    """Convert numpy (image) array to ByteIO stream.
    Useful to stream an image via sockets."""
    img = Image.fromarray(np_data)
    img_memory_file = io.BytesIO()
    img.save(img_memory_file, format)
    return img_memory_file


def memory_file2ndarray(memfile):
    """Load an image stored in the given io.BytesIO memory file memfile."""
    if memfile is None:
        return None
    return np.asarray(Image.open(memfile))


def clip_rect_to_image(rect, img_width, img_height):
    """Ensure that the rectangle is within the image boundaries.
    Explicitly casts all entries to integer.
    :param rect: list/tuple (l,t,w,h)
    :param img_width: int
    :param img_height: int
    """
    li, ti, wi, hi = int(rect[0]), int(rect[1]), int(rect[2]), int(rect[3])
    wi = wi if (li + wi) < img_width else (img_width - max(li, 0) - 1)
    hi = hi if (ti + hi) < img_height else (img_height - max(ti, 0) - 1)
    li = li if li >= 0 else 0
    ti = ti if ti >= 0 else 0
    return [li, ti, wi, hi]


def is_valid_bbox(rect):
    """Left/top must be >= 0, W/H must be > 0"""
    return rect[0] >= 0 and rect[1] >= 0 and rect[2] > 0 and rect[3] > 0


def apply_on_bboxes(image_np, bboxes, func):
    """Applies the function func (which returns a modified image) on each
    bounding box.
    Takes care of proper clipping, roi extraction and copying back the results.
    """
    # Ensure the image is writeable
    image_np = image_np.copy()
    bboxes = [clip_rect_to_image(bb, image_np.shape[1], image_np.shape[0])
            for bb in bboxes]
    bboxes = [b for b in bboxes if is_valid_bbox(b)]
    for bb in bboxes:
        l, t, w, h = bb
        # TODO check single vs multi-channel image
        roi = image_np[t:t+h, l:l+w]
        image_np[t:t+h, l:l+w] = func(roi)
    return image_np


def roi(image_np, rect):
    """Returns the cropped ROI, with rect = [l, t, w, h]."""
    if rect is None or any([r is None for r in rect]):
        return None
    l, t, w, h = rect
    r, b = l+w, t+h
    img_height, img_width = image_np.shape[0], image_np.shape[1]

    # Right/bottom bounds are exclusive; left/top are inclusive
    l = max(0, min(img_width-1, l))
    r = max(0, min(img_width, r))
    t = max(0, min(img_height-1, t))
    b = max(0, min(img_height, b))

    if len(image_np.shape) == 2:
        return image_np[t:b, l:r]
    return image_np[t:b, l:r, :]



import numpy as np
from ..imutils import flip_layers


def test_flip_layers():
    # Single channel image
    x = np.random.rand(3, 3)
    xf = flip_layers(x)
    assert np.all(x == xf)

    # rgb to bgr conversion
    x = np.random.rand(16, 32, 3)
    xf = flip_layers(x)
    for c in range(x.shape[2]):
        assert np.all(x[:, :, c] == xf[:, :, x.shape[2]-c-1])

    # rgbA should be flipped to bgrA
    x = np.random.rand(17, 5, 4)
    xf = flip_layers(x)
    for c in range(x.shape[2]):
        complement = 2-c if c < 3 else 3
        assert np.all(x[:, :, c] == xf[:, :, complement])



import numpy as np
from ..imvis import pseudocolor, color_by_id, exemplary_colors
from .. import colormaps


def assert_color_equal(a, b):
    for c in range(3):
        assert a[c] == b[c]


def test_pseudocolor():
    data = np.array([[0, 1, 2], [255, 42, 0]], dtype=np.uint8)
    for cm in [colormaps.colormap_parula_rgb, colormaps.colormap_jet_rgb, \
            colormaps.colormap_magma_rgb, colormaps.colormap_gray]:
        pc = pseudocolor(data, limits=None, color_map=cm)
        assert np.all(pc[0, 0, :] == pc[1, 2, :])
        assert_color_equal(pc[0, 1, :], cm[1])
        assert_color_equal(pc[0, 2, :], cm[2])
        assert_color_equal(pc[1, 0, :], cm[-1])
        assert_color_equal(pc[1, 1, :], cm[42])
        assert_color_equal(pc[1, 2, :], cm[0])

        pc = pseudocolor(data, limits=[0, 0], color_map=cm)
        for r in range(data.shape[0]):
            for c in range(data.shape[1]):
                assert_color_equal(pc[r, c, :], cm[0])


def test_color_by_id():
    assert_color_equal(color_by_id(0), exemplary_colors[0])
    nc = len(exemplary_colors)
    assert_color_equal(color_by_id(nc), exemplary_colors[0])
    assert_color_equal(color_by_id(nc-1), exemplary_colors[-1])
    assert_color_equal(color_by_id(-1), exemplary_colors[-1])
    assert_color_equal(color_by_id(-3), exemplary_colors[-3])



import argparse
import datetime
import numpy as np
import pytest
from ..pyutils import compare_version_strings, slugify, find_first_index, \
    find_last_index, argsort, is_tool, date_str, check_positive_int, \
    check_positive_real


def test_compare_version_strings():
    assert compare_version_strings('0.1', '0.2') < 0
    assert compare_version_strings('0.17', '0.2') > 0
    assert compare_version_strings('1.5', '1.5') == 0
    assert compare_version_strings('1.5.0', '1.4') > 0
    assert compare_version_strings('1.4.9', '1.6') < 0


def test_slugify():
    assert slugify('') == ''
    assert slugify('abs 17#') == 'abs-17'
    assert slugify('  a  1 ') == 'a-1'


def test_find_indices():
    x = [0, 1, 17, -5, 17, 3, 3, 2, 0, -5]
    assert find_first_index(x, 0) == 0
    assert find_first_index(x, 17) == 2
    assert find_first_index(x, 3) == 5

    assert find_last_index(x, 0) == 8
    assert find_last_index(x, 3) == 6
    assert find_last_index(x, -5) == 9
    assert find_last_index(x, 17) == 4
    assert find_last_index(x, 2) == find_first_index(x, 2)


def test_argsort():
    x = ['z', 'foo', 'bar', 'xkcd']
    expected = [2, 1, 3, 0]
    sidx = argsort(x, indices_only=True)
    assert all([sidx[i] == expected[i] for i in range(len(x))])

    sidx, sx = argsort(x, indices_only=False)
    assert all([sidx[i] == expected[i] for i in range(len(x))])
    assert all(sx[i] == x[sidx[i]] for i in range(len(x)))


def test_date_str():
    ts = datetime.datetime.now()
    year = ts.strftime('%Y')
    month = ts.strftime('%m')
    day = ts.strftime('%d')
    hour = ts.strftime('%H')
    minute = ts.strftime('%M')
    sec = ts.strftime('%S')
    assert date_str([], ts) == year
    assert date_str([''], ts) == ''.join([year, month])
    assert date_str(['_-_'], ts) == '_-_'.join([year, month])
    assert date_str(['#', '#'], ts) == '#'.join([year, month, day])
    assert date_str(['!', '!', '!'], ts) == '!'.join([year, month, day, hour])
    assert date_str(['-', '-', '-', '-'], ts) == \
        '-'.join([year, month, day, hour, minute])
    assert date_str(['_', '_', '_', '_', '_'], ts) == \
        '_'.join([year, month, day, hour, minute, sec])
    assert date_str(['', '-', '<#>', ':', ':|:_#'], ts) == \
        '{}{}-{}<#>{}:{}:|:_#{}'.format(year, month, day, hour, minute, sec)


def test_is_tool():
    # TODO check if 'dir' works on Windows/Mac test servers
    assert is_tool('fooblablub') == False
    assert is_tool('dir') == True


def test_check_positive_int():
    for v in [-17, -1000000, 0, -1, -17.42, -0.0001]:
        with pytest.raises(argparse.ArgumentTypeError):
            check_positive_int(v)
    for v in [1, 1.01, 17, 23, 555.777, 10000000.01]:
        check_positive_int(v)


def test_check_positive_real():
    for v in [-17, -1000000, 0, -1.27, -17.42, -0.0001]:
        with pytest.raises(argparse.ArgumentTypeError):
            check_positive_real(v)
    for v in [0.0000001, 1, 1.01, 17, 23, 555.777, 10000000.01]:
        check_positive_real(v)


#!/usr/bin/env python
# coding=utf-8
"""Geometry utilities related to projections, camera calibration, etc."""

import numpy as np
import numbers

from . import pyutils as pu

# Check, if we can use np.matmul()
__has_np_matmul = pu.compare_version_strings(np.version.version, '1.10.0') >= 0

"""Util to use the correct matrix multiplication across different NumPy versions."""
matmul = np.matmul if __has_np_matmul else np.dot


def dot(a,b):
    """Dot product of Dx1 vectors.
    :params a,b: np.array    
    """
    las = len(a.shape)
    lbs = len(b.shape)
    if las != lbs:
        if las == 1:
            a = a.reshape((a.shape[0], 1))
        if lbs == 1:
            b = b.reshape((b.shape[0], 1))
        las = len(a.shape)
        lbs = len(b.shape)

    if las > 1 and lbs > 1:
        assert a.shape == b.shape
        assert a.shape[1] == 1, 'Currently we only support vector dot products'

        sum = 0.0
        for i in range(a.shape[0]):
            sum += a[i] * b[i]
        return sum
    elif las == 1 or lbs == 1:
        return np.dot(a, b)
    else:
        raise ValueError('Parameter shape mismatch')


def apply_transformation(T, pts):
    """Returns T*coords.
    :param T: MxD transformation matrix
    :param pts: DxN or (D-1)xN data points to transform. If dimension is D-1,
        a homogeneous dimension will be added.
    """
    T = T.astype(np.float64)
    pts = pts.astype(np.float64)
    ndim = T.shape[1]
    # Add homogeneous coordinate if necessary
    if pts.shape[0] == ndim-1:
        npts = pts.shape[1]
        pts  = np.row_stack((pts, np.ones((1,npts), np.float64)))

    # Dimension check
    if pts.shape[0] != ndim:
        raise ValueError('Dimensions do not match')

    return matmul(T, pts)


def apply_projection(P, pts):
    """Computes P*pts and returns the result after dehomogenization.
    :param P: MxD projection matrix
    :param pts: DxN or (D-1)xN data points to project. If dimension is D-1, a
        homogeneous dimension will be added.
    :returns: (M-1)xN projected points after dehomogenization.
    """
    return apply_dehomogenization(apply_transformation(P, pts))


def apply_dehomogenization(pts):
    """Computes the dehomogeneous points, i.e. dividing by the last dimension.
    :param pts: DxN points
    :returns: (D-1)xN dehomogeneous points
    """
    return pts[:-1,:] / pts[-1,:]


def shift_points_along_viewing_rays(pts, distance):
    """Shifts points along the viewing rays in a pinhole camera model. Usually
    used with 3xN pts: x and y coordinates will be computed, such that
    z=distance; distance can be a scalar or a 1xN array
    """
    pts = pts.astype(np.float64)
    num_pts = pts.shape[1]
    # Leverage similar triangles
    if isinstance(distance, numbers.Number):
        shifted = np.row_stack((pts[:-1,:] / pts[-1,:] * distance,
            np.array([distance]*num_pts, dtype=np.float64)))
    else:
        if distance.shape[1] != num_pts:
            raise ValueError('Number of pts and distances must match: {} points, but {} distances!'.format(num_pts, distance.shape[1]))
        distance = distance.astype(np.float64)
        shifted = np.row_stack((np.multiply(pts[:-1,:] / pts[-1,:], distance),
            distance))
    return shifted

def P_from_K_R_t(K, R, t):
    """Returns the 3x4 projection matrix P = K [R | t]."""
    K = K.astype(np.float64)
    R = R.astype(np.float64)
    t = t.astype(np.float64)
    return matmul(K, np.column_stack((R, t)))


def project_world_to_image_K_Rt(K, Rt, world_pts):
    """Projects 3D world points onto the image plane, assuming a calibrated
    camera and rectified inputs.
    """
    K = K.astype(np.float64)
    Rt = Rt.astype(np.float64)
    P = matmul(K, Rt)
    return apply_projection(P, world_pts)


def project_world_to_image_K_R_t(K, R, t, world_pts):
    """Project 3D world points onto the image plane, assuming you already
    calibrated your cameras and rectified your images.

    Don't confuse C (3D position vector of the camera's optical center) and
    t = -RC !!! If you have C, you'll want to use the other
    project_world_to_image().
    """
    R = R.astype(np.float64)
    t = t.astype(np.float64)
    Rt = np.column_stack((R, t))
    return project_world_to_image_K_Rt(K, Rt, world_pts)


# TODO test
def project_world_to_image_K_R_C(K, R, C, world_pts):
    """Project 3D world points onto the image plane, assuming you already
    calibrated your cameras and rectified your images.

    Don't confuse C (3D position vector of the camera's optical center) and
    t = -RC !!! If you have t, you'll want to use the other
    project_world_to_image().
    """
    R = R.astype(np.float64)
    C = C.astype(np.float64)
    Rt = np.column_stack((R, -matmul(R,C)))
    return project_world_to_image_K_Rt(K, Rt, world_pts)


def project_world_to_image_with_distortion_K_Rt(K, Rt, dist_coeff, coords):
    """Project 3D world points onto the image plane, compensating for the lens
    distortion.
    Note that Rt = [R, -RC], where C is the position vector of the optical
    center w.r.t. the world/reference coordinate system! Thus, t=-RC.

    See http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html
    for notes on the camera model, how the distortion works, etc.
    """
    K = K.astype(np.float64)
    Rt = Rt.astype(np.float64)
    dist_coeff = dist_coeff.astype(np.float64)
    coords = coords.astype(np.float64)

    # Transform points from the world/reference coordinate system to the camera
    # coordinate system:
    coords_cam = apply_transformation(Rt, coords)

    # Perform the "normalized pinhole projection", i.e. shift the points along
    # the viewing rays to the "image plane". This is done via similar triangles:
    coords_normalized = coords_cam[0:2,:] / coords_cam[2,:]

    # Apply distortion model, since we use imperfect optics:
    kappa1, kappa2, rho1, rho2, kappa3 = dist_coeff
    r2 = np.sum(np.power(coords_normalized, 2), axis=0) # squared radius, r^2 = x^2 + y^2
    r4 = np.multiply(r2, r2)
    r6 = np.multiply(r2, r4)

    rd = kappa1 * r2 + kappa2 * r4 + kappa3 * r6
    radial_distortion = np.multiply(coords_normalized, np.row_stack((rd,rd)))
    xy = np.multiply(coords_normalized[0,:], coords_normalized[1,:])
    x2 = np.power(coords_normalized[0,:],2)
    y2 = np.power(coords_normalized[1,:],2)
    tdx = 2.0 * rho1 * xy + rho2 * (r2 + 2.0 * x2)
    tdy = rho1 * (r2 + 2.0 * y2) + 2.0 * rho2 * xy
    tangential_distortion = np.row_stack((tdx, tdy))
    coords_corrected = coords_normalized + radial_distortion + tangential_distortion

    # Finally, project onto image:
    return apply_projection(K, coords_corrected)


# TODO this is not tested
def project_world_to_image_with_distortion_K_R_C(K, R, C, dist_coeff, coords):
    """Project 3D world points onto the image plane, compensating for the lens
    distortion. See project_world_to_image_with_distortion_K_Rt()
    """
    R = R.astype(np.float64)
    C = C.astype(np.float64)
    Rt = np.column_stack((R, -matmul(R,C)))
    return project_world_to_image_with_distortion_K_Rt(K, Rt, dist_coeff, coords)


# TODO test
def project_world_to_image_with_distortion_K_R_t(K, R, t, dist_coeff, coords):
    """Project 3D world points onto the image plane, compensating for the lens
    distortion. See project_world_to_image_with_distortion_K_Rt()
    """
    R = R.astype(np.float64)
    t = t.astype(np.float64)
    Rt = np.column_stack((R, t))
    return project_world_to_image_with_distortion_K_Rt(K, Rt, dist_coeff, coords)


def normalize_image_coordinates(K, coords):
    """Compute the normalized coordinates of the given pixel coordinates."""
    invK = np.linalg.inv(K.astype(np.float64))
    return apply_transformation(invK, coords)


def normalize_image_coordinates_with_distortion(K, dist_coeff, pixel_coords):
    """Computes the normalized pinhole projection coordinates of the given pixel
    coordinates and compensates for the lens distortion.
    Basically, it computes:
      x_dist = K^-1 * pixel_coords
      x_norm = compensate_distortion(x_dist)
    """
    # There's no closed form (or even nice) algebraic inverse for the distorted
    # projection. This code is ported from Bouguet's MATLAB toolbox normalize.m
    # We also use his variable names and naming conventions
    pixel_coords = pixel_coords.astype(np.float64)
    K = K.astype(np.float64)
    alpha_c = K[0,1] / K[0,0]  # In Bouguet's toolbox, K[0,:] = [f_x, alpha_c * f_x, cc_x], alpha_c is the skew coefficient
    fc = (K[0,0], K[1,1])
    cc = (K[0,2], K[1,2])
    dist_coeff = dist_coeff.astype(np.float64)

    # Subtract principal point, and divide by the focal length:
    xdx = (pixel_coords[0,:] - cc[0]) / fc[0]
    xdy = (pixel_coords[1,:] - cc[1]) / fc[1]

    # Second, undo skew
    xdx = xdx - alpha_c * xdy

    x_distort = np.row_stack((xdx, xdy))

	# Third, compensate for lens distortion:
    xn = comp_distortion_oulu(x_distort, dist_coeff)
    npts = xn.shape[1]
    return np.row_stack((xn, np.ones((1,npts), dtype=np.float64)))


def comp_distortion_oulu(xd, k):
    """Compensates for radial and tangential distortion. Model From Oulu
    university. This code is a Python port from Bouguet's toolbox, namely
    comp_distortion_oulu.m. We use the same/similar variable names."""
    k1 = k[0]
    k2 = k[1]
    k3 = k[4]
    p1 = k[2]
    p2 = k[3]

    # Initial guess
    x = xd
    num_pts = xd.shape[1]
    for _ in range(20):
        r_2 = np.sum(np.power(x, 2), axis=0)
        # r_2 = sum(x.^2);
        k_radial = np.reshape(1.0 + k1 * r_2 + k2 * np.power(r_2, 2) + k3 * np.power(r_2,3), (1,num_pts))
        # k_radial =  1 + k1 * r_2 + k2 * r_2.^2 + k3 * r_2.^3;
        _xy = np.multiply(x[0,:], x[1,:])
        dxx = 2.0 * p1 * _xy +  p2 * (r_2 + 2.0 * np.power(x[0,:],2))
        dxy = p1 * (r_2 + 2.0 * np.power(x[1,:],2)) + 2.0 * p2 * _xy
        delta_x = np.row_stack((dxx, dxy))

        # Compute relative change (to enable early termination)
        magnitude_delta2 = np.sum(np.power(delta_x, 2), axis=0)
        magnitude_x2 = np.sum(np.power(x, 2), axis=0)
        change = np.divide(magnitude_delta2, magnitude_x2)

        # delta_x = [ ...
        #     2   * p1 * x(1,:).*x(2,:) + p2 * (r_2 + 2   * x(1,:).^2);...
        #     p1 * (r_2 + 2   * x(2,:).^2         ) + 2   * p2 * x(1,:).*x(2,:)];
        x = np.divide(xd - delta_x, np.row_stack((k_radial, k_radial)))
        # x = (xd - delta_x)./(ones(2,1)*k_radial);
        # Stop early if there is no major improvement:
        if np.max(change) < 1e-5:
            break
    return x


def get_projection_matrix(K, R, t):
    """Returns the 3x4 camera projection matrix: P = K * [R | t]"""
    K = K.astype(np.float64)
    R = R.astype(np.float64)
    t = t.astype(np.float64)
    return matmul(K, np.column_stack((R, t)))


def get_groundplane_to_image_homography(P):
    """Given the 3x4 camera projection matrix P, returns the homography
    mapping ground plane points onto the image plane."""
    P = P.astype(np.float64)
    return P[:, [0,1,3]]


def get_image_to_groundplane_homography(P):
    """Given the 3x4 camera projection matrix P, returns the homography
    mapping image plane points onto the ground plane."""
    return np.linalg.inv(get_groundplane_to_image_homography(P))





import numpy as np
from ..cam_projections import dot

def test_dot():
    x = np.random.rand(3, 1)
    y = np.random.rand(3)
    d = x[0]*y[0] + x[1]*y[1] + x[2]*y[2]
    assert dot(x, y) == d
    assert dot(y, x) == d



#!/usr/bin/env python
# coding=utf-8
"""Handy visualization tools."""

import numpy as np

from . import colormaps
from . import imutils

try:
    # Try to load OpenCV (in case you installed it in your workspace)
    import cv2

    def imshow(img_np, title="Image", flip_channels=False, wait_ms=10):
        """
        Convenience 1-liner to display image and wait for key input.

        :param img_np:    Should be provided as BGR, otherwise use flip_channels=True
        :param title:  Window title
        :param wait_ms: cv2.waitKey() input
        :param flip_channels: if you want to display an RGB image poperly, we need
                            to flip the color channels
        :return: Pressed key or -1, i.e. cv2.waitKey() output
        """
        if flip_channels:
            disp = imutils.flip_layers(img_np)
        else:
            disp = img_np
        cv2.imshow(title, disp)
        return cv2.waitKey(wait_ms)
except:
    from PIL import Image

    def imshow(img_np, title="Image", flip_channels=False, **kwargs):
        """
        Convenience 1-liner to display image. This implementation uses
        PIL (which uses your OS default viewer).
        'kwargs' will silently be ignored and are only provided to be
        compatible with the OpenCV-based 'imshow' version (which will be
        loaded in case 'cv2' is installed in your python workspace).

        Note that the window usually doesn't block!

        :param img:    Should be provided as BGR, otherwise use flip_channels=True
        :param title:  Window title
        :param flip_channels: if you want to display an RGB image poperly, we need
                            to flip the color channels
        :return: -1 for compatibility reasons (the same return value as if you
                        used the OpenCV-based version and there was no key press)
        """
        if flip_channels:
            disp = imutils.flip_layers(img_np)
        else:
            disp = img_np
        im = Image.fromarray(disp)
        im.show(title=title)
        return -1


# Colors used to easily distinguish different IDs (e.g. when visualizing
# tracking results).
exemplary_colors = [
    (255,   0,  0),   # red
    (  0, 200,   0),  # light green(ish)
    (  0,   0, 255),  # deep blue
    (230, 230,   0),  # yellow(ish)
    (230,   0, 230),  # magenta(ish)
    (  0, 230, 230),  # cyan(ish)
    (255, 128,   0),  # orange
    (255, 128, 128),  # skin(ish)
    (128,  64,   0),  # brown(ish)
    (160, 160, 160),  # gray(ish)
    (  0, 128, 255),  # light blue
    (153,  77, 204)   # lilac
]


def color_by_id(id, flip_channels=False):
    """Returns a color tuple (rgb) to colorize labels, identities,
    segments, trajectories, etc."""
    col = exemplary_colors[id % len(exemplary_colors)]
    if flip_channels:
        return (col[2], col[1], col[0])
    return col


def pseudocolor(values, limits=[0.0, 1.0], color_map=colormaps.colormap_parula_rgb):
    """
    Return a HxWx3 pseudocolored representation of the input matrix.

    :param values: A single channel, HxW or HxWx1 numpy ndarray.
    :param limits: [min, max] to clip the input values. If limits is None or
        any of min/max is None, the corresponding limits will be computed from
        the input values.
    :param color_map: The color map to be used, see colormaps.py

    :return: a HxWx3 colorized representation.
    """
    # Sanity checks
    if len(values.shape) > 2:
        if values.shape[2] > 1:
            raise ValueError('Input to pseudocoloring must be a single channel data matrix, shaped (H,W) or (H,W,1)!')
        values = values.reshape((values.shape[0], values.shape[1]))

    if limits is None:
        limits = [np.min(values[:]), np.max(values[:])]
    if limits[0] is None:
        limits[0] = np.min(values[:])
    if limits[1] is None:
        limits[1] = np.max(values[:])

    values = values.astype(np.float64)
    lut = np.asarray(color_map)
    interval = (limits[1] - limits[0]) / 255.0
    # Clip values to desired limits
    values[values < limits[0]] = limits[0]
    values[values > limits[1]] = limits[1]
    # Compute lookup values
    if interval > 0:
        lookup_values = np.floor((values - limits[0]) / interval).astype(np.int32)
    else:
        lookup_values = np.zeros(values.shape, dtype=np.int32)
    colorized = lut[lookup_values].astype(np.uint8)
    return colorized


def overlay(img1, img2, weight1, mask1=None):
    """Overlay two images with alpha blending, s.t.
    out = img1 * weight1 + img2 * (1-weight2).
    Optionally, only overlays those parts of the image which are indicated by
    non-zero mask1 pixels: out = blended if mask1 > 0, else img2
    Output dtype will be the same as img2.dtype.
    Only float32, float64 and uint8 are supported.
    """
    if weight1 < 0.0 or weight1 > 1.0:
        raise ValueError('Weight factor must be in [0,1]')

    # Allow overlaying a grayscale image on top of a color image (and vice versa)
    if len(img1.shape) == 2:
        channels1 = 1
    else:
        channels1 = img1.shape[2]

    if len(img2.shape) == 2:
        channels2 = 1
    else:
        channels2 = img2.shape[2]

    if channels1 == channels2 and not (channels1 == 1 or channels1 == 3):
        raise ValueError('Can only extrapolate single channel image to the others dimension')

    if channels1 == 1 and channels2 > 1:
        img1 = np.repeat(img1[:, :, np.newaxis], channels2, axis=2)
    if channels2 == 1 and channels1 > 1:
        img2 = np.repeat(img2[:, :, np.newaxis], channels1, axis=2)

    num_channels = 1 if len(img1.shape) == 2 else img1.shape[2]

    # Convert to float64, [0,1]
    if img1.dtype == np.uint8:
        scale1 = 255.0
    elif img1.dtype in [np.float32, np.float64]:
        scale1 = 1.0
    else:
        raise ValueError('Datatype {} not yet supported'.format(img1.dtype))
    img1 = img1.astype(np.float64) / scale1

    target_dtype = img2.dtype
    if img2.dtype == np.uint8:
        scale2 = 255.0
    elif img2.dtype in [np.float32, np.float64]:
        scale2 = 1.0
    else:
        raise ValueError('Datatype {} not yet supported'.format(img2.dtype))
    img2 = img2.astype(np.float64) / scale2

    if mask1 is None:
        out = weight1 * img1 + (1. - weight1) * img2
    else:
        if num_channels == 1:
            img1 = np.where(np.repeat(mask1[:, :, np.newaxis], num_channels, axis=2) > 0, img1, img2)
        else:
            img1 = np.where(mask1 > 0, img1, img2)
        out = weight1 * img1 + (1. - weight1) * img2
        # out = img2
        # idx = np.where(mask1 > 0)
        # out[idx] = weight1 * img1[idx] + (1. - weight1) * img2[idx]
    return (scale2 * out).astype(target_dtype)



#!/usr/bin/env python
# coding=utf-8
"""Optical flow I/O and visualization."""

import sys
import numpy as np
from . import colormaps


def floread(filename):
    """
    Read optical flow (.flo) files stored in Middlebury format.

    Adapted from https://stackoverflow.com/a/28016469/400948
    """
    if sys.byteorder != 'little':
        raise RuntimeError('Current .flo support requires little-endian architecture!')

    with open(filename, 'rb') as f:
        # Check magic number
        magic = np.fromfile(f, np.float32, count=1)
        if 202021.25 != magic:
            raise ValueError('Invalid magic number in file "%s".' % filename)
        # Next, get the image dimensions
        w = int(np.fromfile(f, np.int32, count=1))
        h = int(np.fromfile(f, np.int32, count=1))
        # Load the data and reshape
        flow = np.fromfile(f, np.float32, count=2*w*h)
        return np.resize(flow, (h, w, 2))


def flosave(filename, flow):
    """Save HxWx2 optical flow to file in Middlebury format."""
    if len(flow.shape) != 3 or flow.shape[2] != 2:
        raise ValueError('Invalid flow shape!')
    # Prepare data
    h, w = flow.shape[0], flow.shape[1]
    data = np.zeros((h, w*2))
    data[:, np.arange(w)*2] = flow[:, :, 0]
    data[:, np.arange(w)*2 + 1] = flow[:, :, 1]

    with open(filename, 'wb') as f:
        # Write magic number
        np.array([202021.25], np.float32).tofile(f)
        # Write dimensions as W,H (!)
        np.array(w).astype(np.int32).tofile(f)
        np.array(h).astype(np.int32).tofile(f)
        # Write actual data
        data.astype(np.float32).tofile(f)


def colorize_flow(
        u, v, return_rgb=True, colorwheel=colormaps.make_flow_color_wheel()):
    """
    Returns the flow color wheel pseudocolorization of the given optical flow.

    This is a port of the C++/MATLAB code from
    https://people.csail.mit.edu/celiu/OpticalFlow, also based on
    https://github.com/tomrunia/OpticalFlow_Visualization.

    :param u: horizontal flow components as HxW np.ndarray
    :param v: vertical flow components as HxW np.ndarray
    :param return_rgb: bool, whether to return RGB or BGR
    :param colorwheel: provide a num_colors-by-3 color numpy nd.array
            which will be used as color wheel
    :param convert_to_bgr: bool, whether to change ordering and
         output BGR instead of RGB
    :return:  HxWx3 uint8 numpy ndarray.
    """
    vis = np.zeros((u.shape[0], u.shape[1], 3), np.uint8)
    ncols = colorwheel.shape[0]

    rad = np.sqrt(np.square(u) + np.square(v))
    a = np.arctan2(-v, -u) / np.pi

    fk = (a+1) / 2*(ncols-1)
    k0 = np.floor(fk).astype(np.int32)
    k1 = k0 + 1
    k1[k1 == ncols] = 0
    f = fk - k0

    for i in range(colorwheel.shape[1]):
        tmp = colorwheel[:, i]
        col0 = tmp[k0] / 255.0
        col1 = tmp[k1] / 255.0
        col = (1-f)*col0 + f*col1

        idx = (rad <= 1)
        # Increase saturation with radius:
        col[idx] = 1 - rad[idx] * (1-col[idx])
        # Out of range:
        col[~idx] = col[~idx] * 0.75
        # BGR or RGB?
        ch_idx = i if return_rgb else 2-i
        vis[:, :, ch_idx] = np.floor(255.0 * col)
    return vis


def flow_to_color(flow, max_val=None, return_rgb=True, epsilon=1e-5):
    """
    Returns the widely used flow visualization.

    This is a port of the C++/MATLAB code from
    https://people.csail.mit.edu/celiu/OpticalFlow, also based on
    https://github.com/tomrunia/OpticalFlow_Visualization.

    :param flow:       stacked u, v flow components as HxWx2 numpy ndarray.
    :param max_val:    float, if not None, flow values will be clipped to
                       the range [0, max_val].
    :param return_rgb: bool, set to False if you work with BGR images.
    :param epsilon:    small value to prevent division-by-zero.
    :return:  HxWx3 uint8 numpy ndarray.
    """
    # Sanity check
    if flow.ndim != 3 or flow.shape[2] != 2:
        raise ValueError('Input flow must be of shape HxWx2!')
    # Clip values
    if max_val is not None:
        flow = np.clip(flow, 0, max_val)

    u = flow[:, :, 0]
    v = flow[:, :, 1]

    # Normalize flow
    rad = np.sqrt(np.square(u) + np.square(v))
    rad_max = np.max(rad)
    u = u / (rad_max + epsilon)
    v = v / (rad_max + epsilon)

    return colorize_flow(u, v, return_rgb)
